<?php

/*
 * @package RS-SITE
 * @copyright (c) 2013 Alexey Glumov aka Rio-Shaman (support@rio-shaman.ru)
 * @license GNU General Public License version 2; see LICENSE.txt
 *
 */
 

/* rs_bbcode.lib.php, v 0.2 2013/10/11 - Парсер bb кодов для движка RS-SITE
 *                                       Работает с визуальным редактором
 *                                       tinyMCE 3.x
 * Копирайт     - (c) 2013
 * Автор        - Rio-Shaman
 * Сайт автора  - http://rio-shaman.ru/
 *
 * Библиотека работает с следующими ББ кодами
 * [b]Bold[/b]
 * [i]Italic[/i]
 * [u]Underline[/u]
 * [code]Some code[/code]
 * [quote]Some quote[/quote]
 * [color=#ff0000]Some color[/color]
 * [url=http://www.moxiecode.com]http://www.moxiecode.com[/url]
 * [img]img/tlogo.png[/img]
 * [left]left[/left]
 * [center]center[/center]
 * [rigth]right[/right]
 *
 * ПРИМЕР ИСПОЛЬЗОВАНИЯ:
 *      ЗАМЕНА ББ КОДОВ НА HTML
 *       $str = "[quote][b]Парсер [i][color=red]bb кодов[/color][/i][/b] для [u]движка[/u] [url=http://rio-shaman.ru/][b]RS-[i]SITE[/i][/b][/url][/quote]";
 *
 *       $rs_bbcode = new rs_bbcode();
 *
 *       $str = $rs_bbcode
 *           ->setText($str)
 *           ->setTag('b')
 *           ->setTag('i')
 *           ->setTag('u')
 *           ->setTag('center')
 *           ->setTag('url')
 *           ->setTag('code')
 *           ->setTag('quote')
 *           ->setTag('color')
 *           ->setTag('img')
 *           ->setTag('left')
 *           ->setTag('right')
 *           ->getHtml()
 *       ;
 *
 *       echo $str;
 *
 *       УДАЛЕНИЕ ББ КОДОВ
 *
 *       $rs_bbcode = new rs_bbcode();
 *
 *       $str = $rs_bbcode
 *           ->setText($str)
 *           ->setTag('b')
 *           ->setTag('i')
 *           ->setTag('u')
 *           ->setTag('center')
 *           ->setTag('url')
 *           ->setTag('code')
 *           ->setTag('quote')
 *           ->setTag('color')
 *           ->setTag('img')
 *           ->setTag('left')
 *           ->setTag('right')
 *           ->getText()
 *       ;
 */

class rs_bbcode
{
    /* 
     *
     * СВОЙСТВА
     *
     */
    
    // шаги поиска. 0 - ищем открытие тега; 1 - ищем закрытие тега; 2 - ищем открытый тег в режиме поиска закрытого тега
    private $step = 0;
    
    // cвойтсва хранит имя искомого тега
    private $saveTag = '';
    
    // дерево элементов
    private $tree    = array();
    
    // cвойства содержит текст сообщения
    private $text = '';
    
    // массив содержит список бб кодов которые предположительно есть в тексте
    private $tag  = array();
    
    // cвойство холостого прохода
    private $blank = 0;
    
    // позиции. начало бб кода (открытый тег) и конец (закрытый)
    private $coordinates        = array();
   
    private $timeCoordinates    = array(
        'open'   => 0,
        'close'  => 0
    );
    
    // свойства содержащие смещение в строчке
    private $cursor   = 0;
    
    // шаблоны определения имена тегов. шаблон в [имя тега][0] является шаблоном для открытого тега, [имя тега][1] - закрытый тег
    private $pattern  = array(
        'b'      => array(
            0    => '/^\[b\]$/si',
            1    => '/^\[\/b\]$/si'
        ),
        'i'      => array(
            0    => '/^\[i\]$/si',
            1    => '/^\[\/i\]$/si'
        ),
        'u'      => array(
            0    => '/^\[u\]$/si',
            1    => '/^\[\/u\]$/si'
        ),
        'code'   => array(
            0    => '/^\[code\]$/si',
            1    => '/^\[\/code\]$/si'
        ),
        'quote'  => array(
            0    => '/^\[quote\]$/si',
            1    => '/^\[\/quote\]$/si'
        ),
        'color'  => array(
            0    => '/^\[color.*?=.*?\]$/si',
            1    => '/^\[\/color\]$/si'
        ),
        'url'    => array(
            0    => '/^\[url.*?=.*?\]$/si',
            1    => '/^\[\/url\]$/si'
        ),
        'img'    => array(
            0    => '/^\[img\]$/si',
            1    => '/^\[\/img\]$/si'
        ),
        'left'   => array(
            0    => '/^\[left\]$/si',
            1    => '/^\[\/left\]$/si'
        ),
        'center' => array(
            0    => '/^\[center\]$/si',
            1    => '/^\[\/center\]$/si'
        ),
        'right'  => array(
            0    => '/^\[right\]$/si',
            1    => '/^\[\/right\]$/si'
        )
    );
    
    // массив содержит запрещенные вложенности
    private $sysBannedTags = array(
        'code'   => array('b', 'i', 'u', 'code', 'quote', 'color', 'url', 'img', 'left', 'center', 'right'),
        'img'    => array('b', 'i', 'u', 'code', 'quote', 'color', 'url', 'img', 'left', 'center', 'right'),
        'url'    => array('code', 'quote', 'url', 'left', 'center', 'right'),
        'left'   => array('code', 'quote', 'url', 'left', 'center', 'right'),
        'center' => array('code', 'quote', 'url', 'left', 'center', 'right'),
        'right'  => array('code', 'quote', 'url', 'left', 'center', 'right'),
        'quote'  => array('code', 'left', 'center', 'right', 'img')
    );




    /* 
     *
     * МЕТОДЫ
     *
     */



    /* СбОР ДАННЫХ */



    // метод устанавливает текст для обработки
    public function setText($text)
    {
        $this->tree = $text;

        return $this;
    }

    // метод состовляющий список бб кодов
    public function setTag($tag)
    {
        $this->tag[] = $tag;
        
        return $this;
    }
    


    /* СТАРТЫ */


    
    // BB CODE -> HTML (старт)
    public function getHtml()
    {
        $this->clearTag();
        $this->clearBlank();
        $this->clearCursor();
        $this->clearTimeCoordinates();
        $this->clearCoordinates();
        $this->clearSaveTage();
        $this->clearStep();
        
        $tree = $this->parse();

        return $this->replaceHtml($tree);
    }
    
    // Удаление бб кодов из текста
    public function getText()
    {
        $this->clearTag();
        $this->clearBlank();
        $this->clearCursor();
        $this->clearTimeCoordinates();
        $this->clearCoordinates();
        $this->clearSaveTage();
        $this->clearStep();
        
        $tree = $this->parse();
        
        return $this->replaceText($tree);
    }



    /* СБОР ДЕРЕВА ЭЛЕМЕНТОВ */
    


    // парсер
    private function parse()
    {
        return $this->searchTag($this->tree);
    }

    // поиск тегов в тексте
    private function searchTag($text)
    {
        $result     = array();
        $this->text = $text;

        // меняем кодировкуу текста
        $this->changeEncoding();

        $length     = strlen($this->text);

        for($i = 0; $i < $length; $i++) {
            
            if($this->text[$i] == '[') {
                $this->cursor = $i;

                $this->definitionTag($this->step);
            }            

            // если курсор достиг конца текста, а режим поиска стоит в положение "ищем закрытый тег"
            // то делаем смещение на +1 от первой верной координаты, и запускаем поиск в режиме
            // "ищем открытый тег"
            if ($i == ($length - 1) AND $this->step == 1) {
                $i = $this->timeCoordinates['open'] + 1;
                
                $this->clearStep();
                $this->clearBlank();
                $this->clearTimeCoordinates();
            }
        }

        $result = $this->shapeSub();

        if (!empty($result))
            foreach ($result as $key => $elemet)
                $result[$key]['sub'] = $this->searchTag($elemet['sub']);
        else
            $result = $this->returnEncoding($this->text);

        return $result;
    }

    // определение тега, запись его координат
    private function definitionTag($step)
    {
        $tag_name = '';

        for($i = $this->cursor; $i >= 0; $i++)
            if (isset($this->text[$i])) {
                $tag_name .= $this->text[$i]; 
                
                if ($this->text[$i] == ']')
                    break;
            }
            else
                return FALSE;

        switch ($step) {
            case '0':
                foreach ($this->pattern as $tag => $pattern)
                    if (preg_match($pattern[$step], $tag_name)) {

                            $this->saveTag = $this->getTagName($tag_name);

                            $this->setTimeCoordinates();
                            $this->changeStep();
                            
                            if ($this->blank > 0)
                                $this->blank--;
                            
                        return TRUE;
                    }
                break;
            case '1':
                if (isset($this->pattern[$this->saveTag][$step])) {
                    if (preg_match($this->pattern[$this->saveTag][$step], $tag_name)) {
                        
                        if ($this->blank > 0) {
                            $this->blank--;
                            return TRUE;
                        }
                        
                        $this->setTimeCoordinates();
                        $this->changeStep('-1');
                        $this->setCoordinates();
                        
                        return TRUE;
                    }
                    
                    $this->changeStep('1');
                    $this->definitionTag($this->step);
                    $this->changeStep('-1');                    
                }
                break;
            case '2':
                if (preg_match($this->pattern[$this->saveTag][0], $tag_name))
                    $this->blank++;

                break;
        }

        return TRUE;
    }

    // формирование главного массива
    private function shapeSub()
    {
        /*
        'element'    => '',
        'attr'       => '',
        'text'       => array(
            '',
            ''
        ),
        'sub'        => ''
        */
        
        $tree = array();
        
        if (!empty($this->coordinates))
            foreach ($this->coordinates as $key => $bbArray) {
                
                // формируем элемент "sub" и "attr"
                $position[0]  = $bbArray['cordinates'][0];
                $position[1]  = ($bbArray['cordinates'][1] - $bbArray['cordinates'][0]) + 1;
                
                $str = substr($this->text, $position[0], $position[1]);
                
                $patter       = array(
                    'search'    => array(
                        0    => '/^\['.$bbArray['element'].'\].*?\[\/'.$bbArray['element'].'\]$/is',
                        1    => '/^\['.$bbArray['element'].'[\s\'\"]{0,1}=[\s\'\"]{0,1}.*?\].*?\[\/'.$bbArray['element'].'\]$/is'
                    ),
                    'save'      => array(
                        0    => '/^\['.$bbArray['element'].'\](.*?)\[\/'.$bbArray['element'].'\]$/is',
                        1    => '/^\['.$bbArray['element'].'.*?=(.*?)\](.*?)\[\/'.$bbArray['element'].'\]$/is'
                    )
                );
                    
                if (preg_match($patter['search'][1], $str)) {
                    preg_match($patter['save'][1], $str, $array);
                    
                    $sub  = $array[2];
                    $attr = $array[1];
                }
                elseif (preg_match($patter['search'][0], $str)) {
                    preg_match($patter['save'][0], $str, $array);
                    
                    $sub  = $array[1];
                    $attr = '';
                }
                
                // END
                
                $text = array();
                
                // формируем текст ДО
                
                $lastCordinates = ($key != 0) ? $this->coordinates[$key - 1]['cordinates'][1] + 1 : 0;
                
                $position[0]  = $lastCordinates;
                $position[1]  = $bbArray['cordinates'][0] - $lastCordinates;
                
                $text[0] = ($bbArray['cordinates'][0] != 0) ? substr($this->text, $position[0], $position[1]) : "";
                
                // формируем текст ПОСЛЕ
                if (!isset($this->coordinates[$key + 1])) {
                    
                    $length = strlen($this->text);
                    
                    $position[0]  = $bbArray['cordinates'][1] + 1;
                    $position[1]  = $length - $bbArray['cordinates'][1];
                    
                    $text[1] = substr($this->text, $position[0], $position[1]);
                }
                else
                    $text[1] = "";
                
                // END

                $tree[] = array(
                    'element'    => $bbArray['element'],
                    'attr'       => $this->returnEncoding($attr),
                    'text'       => array(
                        $this->returnEncoding($text[0]),
                        $this->returnEncoding($text[1])
                    ),
                    'sub'        => $this->returnEncoding($sub)
                );
            }
        
        $this->clearCoordinates();
        $this->clearBlank();
        $this->clearCursor();
        
        return $tree;
    }



    /* СбОР МАССИВА В ТЕКСТ С ОБРАБОТАННЫМИ ББ КОДАМИ */



    // сбор текста с представлением бб кодов как html разметка
    private function replaceHtml($subArray, $bannedTags = array())
    {
        $newBannedTags = array();
        $tag           = array();
        
        if (!is_array($subArray))
            return $subArray;
        
        foreach ($subArray as $key => $sub) {
            switch ($sub['element']) {
                case 'b':
                    $tag[0] = (isset($bannedTags['b'])) ? "[b]"  : '<strong>';
                    $tag[1] = (isset($bannedTags['b'])) ? "[/b]" : '</strong>';
                    break;
                case 'i':
                    $tag[0] = (isset($bannedTags['i'])) ? "[i]"  : '<em>';
                    $tag[1] = (isset($bannedTags['i'])) ? "[/i]" : '</em>';
                    break;
                case 'u':
                    $tag[0] = (isset($bannedTags['u'])) ? "[u]"  : '<span style="text-decoration: underline;">';
                    $tag[1] = (isset($bannedTags['u'])) ? "[/u]" : '</span>';
                    break;
                case 'code':
                    $tag[0] = (isset($bannedTags['code'])) ? "[code]"  : '<div class="code"><pre>';
                    $tag[1] = (isset($bannedTags['code'])) ? "[/code]" : '</pre></div>';
                    break;
                case 'quote':
                    $tag[0] = (isset($bannedTags['quote'])) ? "[quote]"  : '<div class="quote">';
                    $tag[1] = (isset($bannedTags['quote'])) ? "[/quote]" : '</div>';
                    break;
                case 'color':
                    $tag[0] = (isset($bannedTags['color'])) ? "[color=".$sub['attr']."]"  : '<span style="color: '.$sub['attr'].'">';
                    $tag[1] = (isset($bannedTags['color'])) ? "[/color]"                  : '</span>';
                     break;
                case 'url':
                    $tag[0] = (isset($bannedTags['url'])) ? "[url=".$sub['attr']."]"  : '<a rel="nofollow" href="'.$sub['attr'].'">';
                    $tag[1] = (isset($bannedTags['url'])) ? "[/url]"                  : '</a>';
                    break;
                case 'img':
                    $tag[0] = (isset($bannedTags['img'])) ? "[img]"                   : '<img src="'.$GLOBALS['rssite']['cfg']['server_root'].'admin/img/?img=';
                    $tag[1] = (isset($bannedTags['img'])) ? "[/img]"                  : '">';
                    break;
                case 'center':
                    $tag[0] = (isset($bannedTags['center'])) ? "[center]"  : '<p style="text-align: center;">';
                    $tag[1] = (isset($bannedTags['center'])) ? "[/center]" : '</p>';
                    break;
                case 'left':
                    $tag[0] = (isset($bannedTags['left'])) ? "[left]"  : '<p style="text-align: left;">';
                    $tag[1] = (isset($bannedTags['left'])) ? "[/left]" : '</p>';
                    break;
                case 'right':
                    $tag[0] = (isset($bannedTags['right'])) ? "[right]"  : '<p style="text-align: right;">';
                    $tag[1] = (isset($bannedTags['right'])) ? "[/right]" : '</p>';
                    break;
            }
            
            $result[$key] = array($sub['text'][0].$tag[0], $tag[1].$sub['text'][1]);
            
            $newBannedTags = $this->setBannedTag($sub['element'], $bannedTags);
            
            $result[$key][] = (is_array($sub['sub'])) ? $this->replaceHtml($sub['sub'], $newBannedTags) : $sub['sub'];
        }

        foreach($result as $key => $value)
           $result[$key] =  $value[0].$value[2].$value[1];

        $result = implode('', $result);
        
        return $result;
    }
    
    // удаление бб кодов из текста
    private function replaceText($subArray)
    {
        if (!is_array($subArray))
            return $subArray;
        
        foreach ($subArray as $key => $sub) {
            $result[$key]   = array($sub['text'][0], $sub['text'][1]);
            $result[$key][] = (is_array($sub['sub'])) ? $this->replaceText($sub['sub']) : $sub['sub'];
        }
        
        foreach($result as $key => $value)
           $result[$key] =  $value[0].$value[2].$value[1];

        $result = implode('', $result);
        
        return $result;
    }



    /* ВСПОМОГАТЕЛЬНЫЕ МЕТОДЫ */



    // получаем имя тега (из "[b]" получаем "b")
    private function getTagName($tag)
    {
        preg_match("/^\[[\/]{0,1}(.*?)[\]|\s|=]/si", $tag, $name);
        return $name[1];
    }

    // собираем список запрещенных тегов
    private function setBannedTag($tag, $bannedTags)
    {
        if (isset($this->sysBannedTags[$tag]))
            foreach ($this->sysBannedTags[$tag] as $bannedTag)
                $bannedTags[$bannedTag] = $bannedTag;
                    
        return $bannedTags;
    }
    
    // метод временных координат тегов
    private function setTimeCoordinates()
    {
        if ($this->step === 0)
            $this->timeCoordinates['open']  = $this->cursor;

        if ($this->step == 1)
            $this->timeCoordinates['close'] = $this->cursor + strlen($this->saveTag) + 2;
    }
    
    // заполняем свойство координатами тегов
    private function setCoordinates()
    {
        $this->coordinates[] = array(
            'element'    => $this->saveTag,
            'cordinates' => array(
                $this->timeCoordinates['open'],
                $this->timeCoordinates['close']            
            )
        );
        
        $this->clearSaveTage();
        $this->clearTimeCoordinates();
    }
    
    // смена режимов поиска
    private function changeStep($shift = 1)
    {
        $this->step = $this->step + (int)$shift;
    }

    // возвращаем кодировку
    private function returnEncoding($text)
    {
        return iconv('windows-1251', 'UTF-8', $text);
    }
    
    // меняем кодировку
    private function changeEncoding()
    {
        $this->text = iconv('UTF-8', 'windows-1251', $this->text);
    }

    // чистка тегов которые не нужно обрабатывать
    private function clearTag()
    {
        $return = array();

        foreach ($this->tag as $tag)
            if (isset($this->pattern[$tag]))
                $return[$tag] = $this->pattern[$tag];
            
        $this->pattern = $return;
    }
    
    // чистка свойства холостого прохода
    private function clearStep()
    {
        $this->step = 0;
    }
    
    private function clearBlank()
    {
        $this->blank = 0;
    }
    
    // чистка курсора
    private function clearCursor()
    {
        $this->cursor = 0;
    }
    
    // чистка свойства с временными координатами
    private function clearTimeCoordinates()
    {
        $this->timeCoordinates = array(
            'open'   => 0,
            'close'  => 0
        );
    }
 
    // чистка свойства с координатами
    private function clearCoordinates()
    {
        $this->coordinates = array();
    }
    
    // чистим свойства с именем тега
    private function clearSaveTage()
    {
        $this->saveTag = '';
    }

}
?>